import npyodbc
import pytest


def get_running_test_dbs() -> dict[str, npyodbc.Connection]:
    """Get the running database services.

    Returns
    -------
    dict[str, npyodbc.Connection]
        Mapping between the connection name and an npyodbc Connection, if available.
    """
    connections = {
        "sqlite": "Driver=SQLite3;Database=:memory:;Charset=UTF8",
        "postgres": "DRIVER={PostgreSQL Unicode};SERVER=localhost;PORT=5432;UID=postgres_user;PWD=postgres_pwd;DATABASE=test",
        "mysql": "DRIVER={MySQL ODBC 8.0 ANSI Driver};SERVER=localhost;UID=root;PWD=root;DATABASE=test;CHARSET=utf8mb4",
        "mssql_2017": "DRIVER={ODBC Driver 18 for SQL Server};SERVER=localhost,1401;UID=sa;PWD=StrongPassword2017;DATABASE=test;Encrypt=Optional",
        "mssql_2019": "DRIVER={ODBC Driver 18 for SQL Server};SERVER=localhost,1402;UID=sa;PWD=StrongPassword2019;DATABASE=test;Encrypt=Optional",
        "mssql_2022": "DRIVER={ODBC Driver 18 for SQL Server};SERVER=localhost,1403;UID=sa;PWD=StrongPassword2022;DATABASE=test;Encrypt=Optional",
    }
    for db, connection_str in connections.items():
        try:
            npyodbc.connect(connection_str, timeout=1)
            print(f"{db} database service is running.")
        except Exception:
            connections[db] = None
            print(f"{db} database service is not running; tests will be skipped.")
    return connections


DATABASES = get_running_test_dbs()

def pytest_runtest_setup(item: pytest.Item):
    """Check that all tests that have database markers have a running database service.

    Tests marked with databases that are not running are skipped.

    Parameters
    ----------
    item : pytest.Item
        Test invocation item; potentially annotated
    """
    for marker in item.iter_markers():
        if marker.name in DATABASES and DATABASES[marker.name] is None:
            pytest.skip(f"{marker.name} database is not running.")



#!/usr/bin/env python

usage = """\
%(prog)s [options] connection_string

Unit tests for Apache Spark. To use, pass a connection string as the parameter.
The tests will create and drop tables t1 and t2 as necessary.

These run using the version from the 'build' directory, not the version
installed into the Python directories.  You must run python setup.py build
before running the tests.

You can also put the connection string into a tmp/setup.cfg file like so:

  [sparktests]
  connection-string=DSN=Spark

These tests use Simba Spark ODBC driver.
The DSN should be configured with UseNativeQuery=0 to pass the tests.
"""

import sys
import uuid
import unittest
from decimal import Decimal
from testutils import *

_TESTSTR = '0123456789-abcdefghijklmnopqrstuvwxyz-'


def _generate_test_string(length):
    """
    Returns a string of composed of `seed` to make a string `length` characters long.

    To enhance performance, there are 3 ways data is read, based on the length of the value, so
    most data types are tested with 3 lengths.  This function helps us generate the test data.

    We use a recognizable data set instead of a single character to make it less likely that
    "overlap" errors will be hidden and to help us manually identify where a break occurs.
    """
    if length <= len(_TESTSTR):
        return _TESTSTR[:length]

    c = int((length + len(_TESTSTR) - 1) / len(_TESTSTR))
    v = _TESTSTR * c
    return v[:length]


class SparkTestCase(unittest.TestCase):

    INTEGERS = [ -1, 0, 1, 0x7FFFFFFF ]
    BIGINTS  = INTEGERS + [ 0xFFFFFFFF, 0x123456789 ]

    SMALL_READ = 100
    LARGE_READ = 4000

    SMALL_STRING = _generate_test_string(SMALL_READ)
    LARGE_STRING = _generate_test_string(LARGE_READ)
    SMALL_BYTES  = bytes(SMALL_STRING, 'utf-8')
    LARGE_BYTES  = bytes(LARGE_STRING, 'utf-8')

    def __init__(self, connection_string, ansi, method_name):
        unittest.TestCase.__init__(self, method_name)
        self.connection_string = connection_string
        self.ansi = ansi

    def setUp(self):
        self.cnxn   = npyodbc.connect(self.connection_string, ansi=self.ansi, autocommit=True)
        self.cursor = self.cnxn.cursor()

        # I've set my test database to use UTF-8 which seems most popular.
        self.cnxn.setdecoding(npyodbc.SQL_WCHAR, encoding='utf-8')
        self.cnxn.setencoding(encoding='utf-8')

        # As of psql 9.5.04 SQLGetTypeInfo returns absurdly small sizes leading
        # to slow writes.  Override them:
        self.cnxn.maxwrite = 1024 * 1024 * 1024

        for i in range(3):
            try:
                self.cursor.execute("drop table if exists t%d" % i)
            except:
                pass


    def tearDown(self):
        try:
            self.cursor.close()
            self.cnxn.close()
        except:
            # If we've already closed the cursor or connection, exceptions are thrown.
            pass

    def _simpletest(datatype, value):
        # A simple test that can be used for any data type where the Python
        # type we write is also what we expect to receive.
        def _t(self):
            self.cursor.execute('create table t1(value %s)' % datatype)
            self.cursor.execute('insert into t1 values (?)', value)
            result = self.cursor.execute("select value from t1").fetchone()[0]
            self.assertEqual(result, value)
        return _t

    def test_drivers(self):
        p = npyodbc.drivers()
        self.assertTrue(isinstance(p, list))

    def test_datasources(self):
        p = npyodbc.dataSources()
        self.assertTrue(isinstance(p, dict))

    # def test_gettypeinfo(self):
    #     self.cursor.getTypeInfo(npyodbc.SQL_VARCHAR)
    #     cols = [t[0] for t in self.cursor.description]
    #     print('cols:', cols)
    #     for row in self.cursor:
    #         for col,val in zip(cols, row):
    #             print(' ', col, val)

    def test_getinfo_string(self):
        value = self.cnxn.getinfo(npyodbc.SQL_CATALOG_NAME_SEPARATOR)
        self.assertTrue(isinstance(value, str))

    def test_getinfo_bool(self):
        value = self.cnxn.getinfo(npyodbc.SQL_ACCESSIBLE_TABLES)
        self.assertTrue(isinstance(value, bool))

    def test_getinfo_int(self):
        value = self.cnxn.getinfo(npyodbc.SQL_DEFAULT_TXN_ISOLATION)
        self.assertTrue(isinstance(value, int))

    def test_getinfo_smallint(self):
        value = self.cnxn.getinfo(npyodbc.SQL_CONCAT_NULL_BEHAVIOR)
        self.assertTrue(isinstance(value, int))


    def test_negative_float(self):
        value = -200
        self.cursor.execute("create table t1(n float)")
        self.cursor.execute("insert into t1 values (?)", value)
        result  = self.cursor.execute("select n from t1").fetchone()[0]
        self.assertEqual(value, result)


    def _test_strtype(self, sqltype, value, colsize=None, resulttype=None):
        """
        The implementation for string, Unicode, and binary tests.
        """
        assert colsize is None or (value is None or colsize >= len(value))

        if colsize:
            sql = "create table t1(s {}({}))".format(sqltype, colsize)
        else:
            sql = "create table t1(s %s)" % sqltype

        self.cursor.execute(sql)
        self.cursor.execute("insert into t1 values(?)", value)

        result = self.cursor.execute("select * from t1").fetchone()[0]

        if resulttype and type(value) is not resulttype:
            value = resulttype(value)

        self.assertEqual(result, value)

    #
    # VARCHAR
    #

    def test_empty_varchar(self):
        self._test_strtype('varchar', '', self.SMALL_READ)

    def test_null_varchar(self):
        self._test_strtype('varchar', None, self.SMALL_READ)

    def test_large_null_varchar(self):
        # There should not be a difference, but why not find out?
        self._test_strtype('varchar', None, self.LARGE_READ)

    def test_small_varchar(self):
        self._test_strtype('varchar', self.SMALL_STRING, self.SMALL_READ)

    def test_large_varchar(self):
        self._test_strtype('varchar', self.LARGE_STRING, self.LARGE_READ)

    def test_varchar_many(self):
        self.cursor.execute("create table t1(c1 varchar(300), c2 varchar(300), c3 varchar(300))")

        v1 = 'ABCDEFGHIJ' * 30
        v2 = '0123456789' * 30
        v3 = '9876543210' * 30

        self.cursor.execute("insert into t1(c1, c2, c3) values (?,?,?)", v1, v2, v3)
        row = self.cursor.execute("select c1, c2, c3 from t1").fetchone()

        self.assertEqual(v1, row.c1)
        self.assertEqual(v2, row.c2)
        self.assertEqual(v3, row.c3)

    def test_chinese(self):
        v = '我的'
        self.cursor.execute("SELECT '我的' AS name")
        row = self.cursor.fetchone()
        self.assertEqual(row[0], v)

        self.cursor.execute("SELECT '我的' AS name")
        rows = self.cursor.fetchall()
        self.assertEqual(rows[0][0], v)

    for value in INTEGERS:
        name = str(value).replace('.', '_').replace('-', 'neg_')
        locals()['test_int_%s' % name] = _simpletest('int', value)

    for value in BIGINTS:
        name = str(value).replace('.', '_').replace('-', 'neg_')
        locals()['test_bigint_%s' % name] = _simpletest('bigint', value)

    for value in "-1234.56  -1  0  1  1234.56  123456789.21".split():
        name = value.replace('.', '_').replace('-', 'neg_')
        locals()['test_decimal_%s' % name] = _simpletest('decimal(20,6)', Decimal(value))

    for value in "-1234.56  -1  0  1  1234.56  123456789.21".split():
        name = value.replace('.', '_').replace('-', 'neg_')
        locals()['test_numeric_%s' % name] = _simpletest('numeric(20,6)', Decimal(value))

    def test_small_decimal(self):
        value = Decimal('100010')       # (I use this because the ODBC docs tell us how the bytes should look in the C struct)
        self.cursor.execute("create table t1(d numeric(19))")
        self.cursor.execute("insert into t1 values(?)", value)
        v = self.cursor.execute("select * from t1").fetchone()[0]
        self.assertEqual(type(v), Decimal)
        self.assertEqual(v, value)


    def test_small_decimal_scale(self):
        # The same as small_decimal, except with a different scale.  This value exactly matches the ODBC documentation
        # example in the C Data Types appendix.
        value = '1000.10'
        value = Decimal(value)
        self.cursor.execute("create table t1(d numeric(20,6))")
        self.cursor.execute("insert into t1 values(?)", value)
        v = self.cursor.execute("select * from t1").fetchone()[0]
        self.assertEqual(type(v), Decimal)
        self.assertEqual(v, value)


    def test_negative_decimal_scale(self):
        value = Decimal('-10.0010')
        self.cursor.execute("create table t1(d numeric(19,4))")
        self.cursor.execute("insert into t1 values(?)", value)
        v = self.cursor.execute("select * from t1").fetchone()[0]
        self.assertEqual(type(v), Decimal)
        self.assertEqual(v, value)

    def _exec(self):
        self.cursor.execute(self.sql)

    def test_close_cnxn(self):
        """Make sure using a Cursor after closing its connection doesn't crash."""

        self.cursor.execute("create table t1(id integer, s varchar(20))")
        self.cursor.execute("insert into t1 values (?,?)", 1, 'test')
        self.cursor.execute("select * from t1")

        self.cnxn.close()

        # Now that the connection is closed, we expect an exception.  (If the code attempts to use
        # the HSTMT, we'll get an access violation instead.)
        self.sql = "select * from t1"
        self.assertRaises(npyodbc.ProgrammingError, self._exec)

    def test_empty_string(self):
        self.cursor.execute("create table t1(s varchar(20))")
        self.cursor.execute("insert into t1 values(?)", "")

    def test_fixed_str(self):
        value = "testing"
        self.cursor.execute("create table t1(s char(7))")
        self.cursor.execute("insert into t1 values(?)", "testing")
        v = self.cursor.execute("select * from t1").fetchone()[0]
        self.assertEqual(type(v), str)
        self.assertEqual(len(v), len(value)) # If we alloc'd wrong, the test below might work because of an embedded NULL
        self.assertEqual(v, value)

    def test_fetchval(self):
        expected = "test"
        self.cursor.execute("create table t1(s varchar(20))")
        self.cursor.execute("insert into t1 values(?)", expected)
        result = self.cursor.execute("select * from t1").fetchval()
        self.assertEqual(result, expected)

    def test_negative_row_index(self):
        self.cursor.execute("create table t1(s varchar(20))")
        self.cursor.execute("insert into t1 values(?)", "1")
        row = self.cursor.execute("select * from t1").fetchone()
        self.assertEqual(row[0], "1")
        self.assertEqual(row[-1], "1")

    def test_version(self):
        self.assertEqual(3, len(npyodbc.version.split('.'))) # 1.3.1 etc.

    def test_lower_case(self):
        "Ensure npyodbc.lowercase forces returned column names to lowercase."

        # Has to be set before creating the cursor, so we must recreate self.cursor.

        npyodbc.lowercase = True
        self.cursor = self.cnxn.cursor()

        self.cursor.execute("create table t1(Abc int, dEf int)")
        self.cursor.execute("select * from t1")

        names = [ t[0] for t in self.cursor.description ]
        names.sort()

        self.assertEqual(names, [ "abc", "def" ])

        # Put it back so other tests don't fail.
        npyodbc.lowercase = False

    def test_long_column_name(self):
        "ensure super long column names are handled correctly."
        c1 = 'abcdefghij' * 50
        c2 = 'klmnopqrst' * 60
        self.cursor = self.cnxn.cursor()

        self.cursor.execute("create table t1(c1 int, c2 int)")
        self.cursor.execute(f"select c1 as {c1}, c2 as {c2} from t1")

        names = [ t[0] for t in self.cursor.description ]
        names.sort()

        self.assertEqual(names, [ c1, c2 ])

    def test_row_description(self):
        """
        Ensure Cursor.description is accessible as Row.cursor_description.
        """
        self.cursor = self.cnxn.cursor()
        self.cursor.execute("create table t1(a int, b char(3))")
        self.cursor.execute("insert into t1 values(1, 'abc')")

        row = self.cursor.execute("select * from t1").fetchone()
        self.assertEqual(self.cursor.description, row.cursor_description)


    def test_executemany(self):
        self.cursor.execute("create table t1(a int, b varchar(10))")

        params = [ (i, str(i)) for i in range(1, 6) ]

        self.cursor.executemany("insert into t1(a, b) values (?,?)", params)

        count = self.cursor.execute("select count(*) from t1").fetchone()[0]
        self.assertEqual(count, len(params))

        self.cursor.execute("select a, b from t1 order by a")
        rows = self.cursor.fetchall()
        self.assertEqual(count, len(rows))

        for param, row in zip(params, rows):
            self.assertEqual(param[0], row[0])
            self.assertEqual(param[1], row[1])

    def test_fast_executemany(self):

        self.fast_executemany = True

        self.cursor.execute("create table t1(a int, b varchar(10))")

        params = [(i, str(i)) for i in range(1, 6)]

        self.cursor.executemany("insert into t1(a, b) values (?,?)", params)

        count = self.cursor.execute("select count(*) from t1").fetchone()[0]
        self.assertEqual(count, len(params))

        self.cursor.execute("select a, b from t1 order by a")
        rows = self.cursor.fetchall()
        self.assertEqual(count, len(rows))

        for param, row in zip(params, rows):
            self.assertEqual(param[0], row[0])
            self.assertEqual(param[1], row[1])

    def test_executemany_failure(self):
        """
        Ensure that an exception is raised if one query in an executemany fails.
        """
        self.cursor.execute("create table t1(a int, b varchar(10))")

        params = [ (1, 'good'),
                   ('error', 'not an int'),
                   (3, 'good') ]

        self.assertRaises(npyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)


    def test_row_slicing(self):
        self.cursor.execute("create table t1(a int, b int, c int, d int)");
        self.cursor.execute("insert into t1 values(1,2,3,4)")

        row = self.cursor.execute("select * from t1").fetchone()

        result = row[:]
        self.assertTrue(result is row)

        result = row[:-1]
        self.assertEqual(result, (1,2,3))

        result = row[0:4]
        self.assertTrue(result is row)

    def test_row_repr(self):
        self.cursor.execute("create table t1(a int, b int, c int, d int)")
        self.cursor.execute("insert into t1 values(1,2,3,4)")

        row = self.cursor.execute("select * from t1").fetchone()

        result = str(row)
        self.assertEqual(result, "(1, 2, 3, 4)")

        result = str(row[:-1])
        self.assertEqual(result, "(1, 2, 3)")

        result = str(row[:1])
        self.assertEqual(result, "(1,)")

    def test_cnxn_set_attr_before(self):
        # I don't have a getattr right now since I don't have a table telling me what kind of
        # value to expect.  For now just make sure it doesn't crash.
        # From the unixODBC sqlext.h header file.
        SQL_ATTR_PACKET_SIZE = 112
        othercnxn = npyodbc.connect(self.connection_string, attrs_before={ SQL_ATTR_PACKET_SIZE : 1024 * 32 }, autocommit=True)

    def test_cnxn_set_attr(self):
        # I don't have a getattr right now since I don't have a table telling me what kind of
        # value to expect.  For now just make sure it doesn't crash.
        # From the unixODBC sqlext.h header file.
        SQL_ATTR_ACCESS_MODE = 101
        SQL_MODE_READ_ONLY   = 1
        self.cnxn.set_attr(SQL_ATTR_ACCESS_MODE, SQL_MODE_READ_ONLY)

    def test_columns(self):
        self.cursor.execute("create table t1(a int, b varchar(3), `xΏz` decimal(8,2))")

        self.cursor.columns('t1')
        results = {row.column_name: row for row in self.cursor}
        row = results['a']
        assert row.type_name == 'INT', row.type_name
        row = results['b']
        assert row.type_name == 'VARCHAR'
        row = results['xΏz']
        assert row.type_name == 'DECIMAL'

        # Now do the same, but specifically pass in None to one of the keywords.  Old versions
        # were parsing arguments incorrectly and would raise an error.  (This crops up when
        # calling indirectly like columns(*args, **kwargs) which aiodbc does.)

        self.cursor.columns('t1', schema=None, catalog=None)
        results = {row.column_name: row for row in self.cursor}
        row = results['a']
        assert row.type_name == 'INT', row.type_name
        row = results['b']
        assert row.type_name == 'VARCHAR'

    def test_cancel(self):
        # I'm not sure how to reliably cause a hang to cancel, so for now we'll settle with
        # making sure SQLCancel is called correctly.
        self.cursor.execute("select 1")
        self.cursor.cancel()

    def test_emoticons_as_parameter(self):
        # https://github.com/mkleehammer/pyodbc/issues/423
        #
        # When sending a varchar parameter, npyodbc is supposed to set ColumnSize to the number
        # of characters.  Ensure it works even with 4-byte characters.
        #
        # http://www.fileformat.info/info/unicode/char/1f31c/index.htm

        v = "x \U0001F31C z"

        self.cursor.execute("CREATE TABLE t1(s varchar(100))")
        self.cursor.execute("insert into t1 values (?)", v)

        result = self.cursor.execute("select s from t1").fetchone()[0]

        self.assertEqual(result, v)

def main():
    from argparse import ArgumentParser
    parser = ArgumentParser(usage=usage)
    parser.add_argument("-v", "--verbose", action="count", default=0, help="increment test verbosity (can be used multiple times)")
    parser.add_argument("-d", "--debug", action="store_true", default=False, help="print debugging items")
    parser.add_argument("-t", "--test", help="run only the named test")
    parser.add_argument("-a", "--ansi", help="ANSI only", default=False, action="store_true")
    parser.add_argument("conn_str", nargs="*", help="connection string for Spark")

    args = parser.parse_args()

    if len(args.conn_str) > 1:
        parser.error('Only one argument is allowed.  Do you need quotes around the connection string?')

    if not args.conn_str:
        connection_string = load_setup_connection_string('sparktests')

        if not connection_string:
            parser.print_help()
            raise SystemExit()
    else:
        connection_string = args.conn_str[0]

    if args.verbose:
        cnxn = npyodbc.connect(connection_string, ansi=args.ansi)
        print_library_info(cnxn)
        cnxn.close()

    if args.test:
        # Run a single test
        if not args.test.startswith('test_'):
            args.test = 'test_%s' % (args.test)

        s = unittest.TestSuite([ SparkTestCase(connection_string, args.ansi, args.test) ])
    else:
        # Run all tests in the class

        methods = [ m for m in dir(SparkTestCase) if m.startswith('test_') ]
        methods.sort()
        s = unittest.TestSuite([ SparkTestCase(connection_string, args.ansi, m) for m in methods ])

    testRunner = unittest.TextTestRunner(verbosity=args.verbose)
    result = testRunner.run(s)

    return result


if __name__ == '__main__':

    # Add the build directory to the path so we're testing the latest build, not the installed version.

    add_to_path()

    import npyodbc
    sys.exit(0 if main().wasSuccessful() else 1)



import string
import warnings
from typing import Optional

import hypothesis.strategies as st
import npyodbc
import numpy as np
import pytest
from hypothesis import given, reject, settings
from numpy.testing import assert_allclose, assert_array_equal


@pytest.fixture(scope="module")
def connection() -> npyodbc.Connection:
    """Return a connection for a SQLite3 database."""
    return npyodbc.connect("Driver=SQLite3;Database=:memory:;Charset=UTF8")


@pytest.fixture(scope="module")
def cursor(connection) -> npyodbc.Cursor:
    """Return a cursor for a SQLite3 database."""
    return connection.cursor()


def cleanup(cursor: npyodbc.Cursor):
    """Drop all tables from the database.

    Use this function to clean up the database at the end of each test rather
    than a fixture because hypothesis doesn't work with test-scoped fixtures.

    Parameters
    ----------
    cursor : npyodbc.Cursor
        Cursor for the database to clean up
    """
    for table in cursor.execute('SELECT tbl_name FROM sqlite_schema;').fetchall():
        cursor.execute(f'DROP TABLE {table[0]};')
    assert len(cursor.execute('SELECT tbl_name FROM sqlite_schema;').fetchall()) == 0


def assert_result_close(a: np.ndarray, b: np.ndarray, dtype: Optional[np.dtype] = None):
    """Assert that a is close to b.

    If a is a float dtype, use assert_allclose; otherwise we check for exact equality.

    Parameters
    ----------
    a : np.ndarray
        An array to check
    b : np.ndarray
        An array to check
    dtype : Optional[np.dtype]
        Dtype of the arrays; used to provide relative tolerance, if provided
    """
    if a.dtype.kind == np.dtype(float).kind:
        if dtype is not None:
            assert_allclose(a, b, atol=np.finfo(dtype).resolution)
        else:
            assert_allclose(a, b)
    else:
        assert_array_equal(a, b)


@pytest.mark.sqlite
def test_cursor_fetchdictarray_method_exists():
    """Test that npyodbc.Cursor.fetchdictarray exists."""
    assert hasattr(npyodbc.Cursor, 'fetchdictarray')


@pytest.mark.sqlite
@settings(deadline=None)
@given(
    st.lists(
        st.tuples(
            st.text(alphabet=string.printable, max_size=10),
            st.integers(min_value=-(2**16)//2, max_value=(2**16)//2 - 1),
            st.floats(allow_nan=False, allow_infinity=False, min_value=-1.79e308, max_value=1.79e308),
        ),
        min_size=1,
        max_size=100,
    )
)
@pytest.mark.parametrize(
    ("target_dtypes"),
    [
        {'a': 'U', 'b': 'i8', 'c': 'f8'},
        {},
    ]
)
def test_fetchdictarray(cursor, target_dtypes, values):
    """Test that fetchdictarray returns the expected values from the database.

    - Text is restricted to printable characters because that's tested separately.
    - Integers are restricted to the 32-bit range (-2**31, 2**31 - 1) because otherwise
      the values get wrapped in the database.
    - Floats are restricted to non-nan, non-inf values between (-1.79e308, 1.79e308),
      which is the range covered by a double precision (64-bit) float. If you try to
      store values larger than this in a REAL column, it gets converted to +/- inf.
    """
    cursor.execute('DROP TABLE IF EXISTS t1;')
    cursor.execute('CREATE TABLE t1(a text, b int, c real);')

    for value in values:
        cursor.execute('INSERT INTO t1 values(?,?,?);', *value)

    inserted = cursor.execute('SELECT * from t1;').fetchdictarray(
        target_dtypes=target_dtypes
    )

    expected = {col[0]: np.array(arr) for col, arr in zip(cursor.description, zip(*values))}
    for col, value in expected.items():
        assert_result_close(inserted[col], value, target_dtypes.get(col))

    cleanup(cursor)


@pytest.mark.sqlite
@pytest.mark.parametrize(
    ("target_dtypes"),
    [
        {'a': 'U', 'b': 'i8', 'c': 'f8'},
        {},
    ]
)
def test_fetchdictarray_unicode(cursor, target_dtypes):
    """Test that fetchdictarray correctly a specific unicode string test case."""
    cursor.execute('DROP TABLE IF EXISTS t1;')
    cursor.execute('CREATE TABLE t1(a text, b int);')
    cursor.execute('INSERT INTO t1 values(?,?);', '𐀀', 0)

    inserted = cursor.execute('SELECT * from t1;').fetchdictarray(
        target_dtypes=target_dtypes
    )
    expected = {'a': np.array(['𐀀']), 'b': np.array([0])}

    for key, value in expected.items():
        assert_array_equal(inserted[key], value)

    cleanup(cursor)


@pytest.mark.sqlite
@settings(deadline=None)
@given(
    st.lists(
        st.tuples(
            st.text(max_size=30).filter(lambda x: x != '\U0010d800'),
            st.integers(min_value=-(2**32)//2, max_value=(2**32)//2 - 1),
        ),
        min_size=1,
        max_size=100,
    ),
)
@pytest.mark.parametrize(
    ("target_dtypes"),
    [
        {'a': 'U', 'b': 'i8'},
        {},
    ]
)
def test_fetchdictarray_unicode_values(cursor, target_dtypes, values):
    """Test that fetchdictarray correctly handles arrays of unicode strings."""
    cursor.execute('DROP TABLE IF EXISTS t1;')
    cursor.execute('CREATE TABLE t1(a text, b int);')

    for text, val in values:
        cursor.execute('INSERT INTO t1 values(?,?);', text, val)

    cursor.execute('SELECT * from t1;')
    inserted = cursor.fetchdictarray(target_dtypes=target_dtypes)

    try:
        cursor.execute('SELECT * from t1;')
        expected_text, expected_ints = zip(*cursor.fetchall())
    except UnicodeDecodeError:
        warnings.warn(
            "Pyodbc itself failed to decode UTF16, but npyodbc did not.\n"
            f"values: {values}",
            stacklevel=1,
        )
        reject()

    expected = {
        'a': np.array(expected_text),
        'b': np.array(expected_ints),
    }

    for key, value in expected.items():
        assert_array_equal(inserted[key], value)

    cleanup(cursor)


@pytest.mark.sqlite
@pytest.mark.parametrize(
    ('sql_type'),
    [
        ('BINARY'),
        ('VARBINARY'),
        ('LONGVARBINARY'),
    ]
)
def test_fetchdictarray_binary(cursor, sql_type):
    """Test that fetchdictarray can retrieve binary, varbinary, and longvarbinary columns."""
    binary = [
        b"foo",
        b"bar",
        b"baz",
        b"quux",
    ]
    ints = [1, 2, 3, 4]

    # Need to specify max element size for binary/varbinary/longvarbinary col.
    elsize = 4
    cursor.execute('DROP TABLE IF EXISTS t1;')
    cursor.execute(f'CREATE TABLE t1(a {sql_type}({elsize}), b int);')

    for bval, ival in zip(binary, ints):
        cursor.execute('INSERT INTO t1 values(?,?);', bval, ival)

    cursor.execute('SELECT * from t1;')

    # Cast column 'a' to a null-terminated bytes dtype
    fda_result = cursor.fetchdictarray(target_dtypes={'a': "S", 'b': 'i8'})

    assert_result_close(
        fda_result['a'], np.array(binary, dtype='S4')
    )
    assert_result_close(
        fda_result['b'], np.array(ints, dtype='i8')
    )

    cleanup(cursor)


@pytest.mark.sqlite
@pytest.mark.parametrize(
    ('sql_type'),
    [
        ('BINARY'),
        ('VARBINARY'),
        ('LONGVARBINARY'),
    ]
)
@pytest.mark.parametrize(
    ('dtype'),
    [
        '?', # boolean
        'i32', # (signed) integer
        'u16', # unsigned integer
        'f32', # floating-point
        'c64', # complex-floating point
        'm', # timedelta
        'M', # datetime
        'O', # (Python) objects
        'V', # raw data (void)
    ]
)
def test_fetchdictarray_binary_bad_type_coercion(cursor, sql_type, dtype):
    """Test that fetchdictarray fails when coercing binary data to nonsensical types."""
    binary = [
        b"foo",
        b"bar",
        b"baz",
        b"quux",
    ]
    ints = [1, 2, 3, 4]

    # Need to specify max element size for binary/varbinary/longvarbinary col. Since this is
    # sqlite, blobs are stored in hexadecimal format, so the length of the stored data is
    # not the same as the length of the bytestrings above. Let's just make it 16 bytes, which
    # is enough to hold the 4 characters in the largest bytestring, plus the extra three
    # added by sqlite, as explained below.
    elsize = 16
    cursor.execute('DROP TABLE IF EXISTS t1;')
    cursor.execute(f'CREATE TABLE t1(a {sql_type}({elsize}), b int);')

    for bval, ival in zip(binary, ints):
        cursor.execute('INSERT INTO t1 values(?,?);', bval, ival)

    cursor.execute('SELECT * from t1;')

    # Cast column 'a' to the requested dtype
    with pytest.raises((SystemError, TypeError)):
        cursor.fetchdictarray(target_dtypes={'a': dtype, 'b': 'i8'})

    cleanup(cursor)



@pytest.mark.sqlite
@settings(deadline=None)
@given(
    st.lists(
        st.tuples(
            st.integers(min_value=np.iinfo('i1').min, max_value=np.iinfo('i1').max),
            st.floats(allow_nan=False, allow_infinity=False, min_value=np.finfo('f8').min, max_value=np.finfo('f8').max),
        ),
        min_size=1,
        max_size=100,
    )
)
@pytest.mark.parametrize(
    ("int_dtype"),
    [
        'i1',
        'i2',
        'i4',
        'i8',
    ]
)
@pytest.mark.parametrize(
    ("float_dtype"),
    [
        'f8', # sqlite only stores 8-byte floating point numbers in REAL format
    ]
)
def test_fetchdictarray_coercion(cursor, int_dtype, float_dtype, values):
    """Test that fetchdictarray returns the expected values from the database when coerced.

    - Integers are restricted to the 8-bit range [-128, 127], which is what fits
      in all int_dtypes.
    - Floats are restricted to non-nan, non-inf values in the 8-byte range
    """
    cursor.execute('DROP TABLE IF EXISTS t1;')
    cursor.execute('CREATE TABLE t1(a int, b real);')

    rounded = []
    ints = []
    for (intval, floatval) in values:
        rounded_float = np.round(
            floatval, decimals=np.finfo(float_dtype).precision
        )

        ints.append(intval)
        rounded.append(rounded_float)
        cursor.execute("INSERT INTO t1 values(?,?);", intval, rounded_float)

    inserted = cursor.execute('SELECT * from t1;').fetchdictarray(
        target_dtypes={'a': int_dtype, 'b': float_dtype}
    )

    assert_array_equal(
        np.array(ints), inserted['a']
    )

    assert_result_close(
        np.array(rounded), inserted['b'], dtype=float_dtype
    )

    cleanup(cursor)
